#include <stdio.h>
#include <assert.h>
#include "parti.h"

/*================================================================
 *
 *  Function   : enter_hash ()
 *  Visibility : Internal 
 *  
 *  Purpose    : Having dereferenced the new elements, we need
 * 			   	 to put them in the hash table
 *  
 *  Parameters : all are in_parameters. hash_table is modified
 *
 *  Returns    : nothing
 * 
 *  Calls      : insert_hash_element() ,  malloc()
 *
 *  Called by  : hash_input()
 *
 *================================================================*/
 
void enter_hash(data_set, data_size,  hash_table, stamp, supermask, 
	                 proc, local, my_size)

int        *data_set;                 /* the input array         */
int         data_size;                /* size of input array     */
HASHSTR    *hash_table;               /* pointer to hash_table   */
int        *stamp;
int         supermask;
int        *proc;
int        *local;
int         my_size;
{

int       i,j, k, key, b, found, 
	  ptr, stmp, no_buckets, tunique, munique, tmask;
int       new_stamp;
int       *chain,  *stamplist , *offset;
HASHCELL  *tmp, *hashcell , *addr;
 

   assert( hash_table  != NULL );
   assert( hash_table->no_stamps  <= 31 );
   assert( hash_table->buckets != NULL);

   if ( supermask == NO_MASK )   {
       assert( hash_table->no_stamps == 0 );
       hash_table->first_offset =   ptr      = my_size;
       hash_table->first_stamp  =  new_stamp = FIRST_STAMP ;
       hash_table->start_offset[new_stamp] = ptr; 
   }
   else {
       stmp      = biggest_stamp_in_mask( supermask );
       ptr       = hash_table->start_offset[stmp] +  
		   hash_table->area_stamp[stmp] ;
       if (  (*stamp ) == NO_STAMP ) {
          new_stamp =  biggest_stamp_in_mask( hash_table->stamp_mask );
       }
       else  {
         new_stamp = 1+biggest_stamp_in_mask( supermask );
         hash_table->start_offset[new_stamp] = ptr; 
       }
   }
   
   if ((*stamp ) != NO_STAMP )   (hash_table->no_stamps)++;
   no_buckets = hash_table->no_buckets ;
   tmask      = make_mask( new_stamp );
   (hash_table->stamp_mask) |=  tmask ;

   tunique    =  munique  =  0;
   for( i=0; i< data_size; i++ ) {
      if ( proc[i] ==  PARTI_mynode ) {
         data_set[i] = local[i];
         continue;
      }
      assert( data_set[i] >= 0 );
      key   = data_set[i];
      b     = key  % no_buckets ;
      addr  = &(hash_table->buckets[b]);
      tmp   = hashcell = &(hash_table->buckets[b]);
      found = 0;

      for( ; hashcell != NULL && !(found) ;
             tmp = hashcell, hashcell = hashcell->next ) {
        chain     = hashcell->key;
        offset    = hashcell->offset;
        stamplist = hashcell->stamp;
        for ( j=0; j<hashcell->tag; j++ ) {      /* SEARCH FOR KEY */
          if ( chain[j] == key ) {
            if ( !( stamplist[j] & tmask )  ) {
               stamplist[j] = (stamplist[j] | tmask) ;
               tunique ++ ;
               if ((supermask == NO_MASK) ||
                   ((supermask != NO_MASK) && (!(stamplist[j] & supermask)))) {
                     munique ++ ;                     /*..... QQQQ */
                     offset[j] = ptr;
                     ptr++;
               }
             }
             data_set[i] = offset[j];
             found = 1;
             break;
          }
        }
      }

     /*  KEY WAS NOT FOUND. It MUST BE INSERTED IN TABLE */

     if (  ! found ) {
       tunique ++ ;  munique ++;
       if ( tmp->tag < CHAINSIZE ) {
         j = tmp->tag;
         tmp->tag = j  + 1 ;
         tmp->key[j]   = key;
         tmp->proc[j]  = proc[i];
         tmp->local[j] = local[i];
         tmp->offset[j]= ptr;
         tmp->stamp[j] = tmask ;
         data_set[i] = ptr;
         ptr++;
       }
       else if ( tmp->tag >= CHAINSIZE ) {
         tmp->next       = ( HASHCELL *)malloc( sizeof(HASHCELL));
         assert( tmp->next != NULL );
         tmp->next->next = NULL;
         tmp->next->tag  = 0;
         tmp = tmp->next;
         for( k=0; k < CHAINSIZE; tmp->key[k]=-2, k++) ;
         for( k=0; k < CHAINSIZE; tmp->stamp[k]=0, k++) ;
         j = 0;
         tmp->key[j]   = key;
         tmp->proc[j]  = proc[i];
         tmp->local[j] = local[i];
         tmp->offset[j]= ptr;
         tmp->stamp[j] = tmask ;
         tmp->tag = j + 1 ;
         data_set[i] = tmp->offset[j];
         ptr++;
       }
     }
   }        /* big for loop */

   if ( (*stamp) == NO_STAMP ) {
      hash_table->size_stamp[new_stamp] += munique  ;
      hash_table->area_stamp[new_stamp] += munique ;
   }
   else  {
      hash_table->size_stamp[new_stamp] = tunique ;
      hash_table->area_stamp[new_stamp] = munique ;
   }
   (*stamp) = new_stamp ;

#ifdef  DEBUG
PARTI_gsync();
#endif
}

/*======================================================================*/
/* 
 * Enter elements to replicated hash table 
 *
 *	Note: The replication factor of translation table must be 1.0.
 *	      The type of hash table must be REPHASH.
 */

void 
enter_rep_hash(trans_table, data_set, data_size, hash_table, 
						stamp, supermask, my_size)
TTABLE     *trans_table;		/* translation table */
int        *data_set;                 /* the input array         */
int         data_size;                /* size of input array     */
HASHSTR    *hash_table;               /* pointer to hash_table   */
int        *stamp;
int         supermask;
int         my_size;

{

int       i,j, k, key, b, found, 
	  ptr, stmp, no_buckets, tunique, munique, tmask;
int       new_stamp, nstamp; 
HASHCELL  *tmp, *hashcell , *addr;
 

   assert( hash_table  != NULL );
   assert( hash_table->no_stamps  <= 31 );

   if ( supermask == NO_MASK )   {
       assert( hash_table->no_stamps == 0 );
       hash_table->first_offset =   ptr      = my_size;
       hash_table->first_stamp  =  new_stamp = FIRST_STAMP ;
       hash_table->start_offset[new_stamp] = ptr; 
   }
   else {
       stmp      = biggest_stamp_in_mask( supermask );
       ptr       = hash_table->start_offset[stmp] +  
		   hash_table->area_stamp[stmp] ;
       if (  (*stamp ) == NO_STAMP ) {
          new_stamp =  biggest_stamp_in_mask( hash_table->stamp_mask );
       }
       else  {
         new_stamp = 1+biggest_stamp_in_mask( supermask);
         hash_table->start_offset[new_stamp] = ptr; 
       }
   }
   
   if ((*stamp ) != NO_STAMP )   (hash_table->no_stamps)++;
   no_buckets = hash_table->no_buckets ;
   tmask      = make_mask( new_stamp );
   (hash_table->stamp_mask) |=  tmask ;

   tunique    =  munique  =  0;
   for ( i=0; i< data_size; i++ ) {
      key = (data_set[i] >= 0)? data_set[i]: -data_set[i];

      /* check if the element is local */
      if ( trans_table->tabptr->Lproc[key] ==  PARTI_mynode ) {
	 /* local data */
	 data_set[i] = (data_set[i] >= 0)
		     ? trans_table->tabptr->Loffset[key]
		     : - trans_table->tabptr->Loffset[key];
	 continue;
      }
 
      /* the element is not local */
      /* check if the element is already seen */
      nstamp = hash_table->cells[key].stamp;
      if ( !( nstamp & tmask )  ) {
	 /* the element has not been seen */
         hash_table->cells[key].stamp = nstamp | tmask;
         tunique ++ ;
         if ((supermask == NO_MASK) ||
	     ((supermask != NO_MASK) && (!(nstamp & supermask)))) {
            munique ++ ;                     /*..... QQQQ */
            hash_table->cells[key].offset = ptr++; 
         }
      }
      data_set[i] = (data_set[i] >= 0)
		  ? hash_table->cells[key].offset
		  : - hash_table->cells[key].offset;
   }

   if ( (*stamp) == NO_STAMP ) {
      hash_table->size_stamp[new_stamp] += munique  ;
      hash_table->area_stamp[new_stamp] += munique ;
   }
   else  {
      hash_table->size_stamp[new_stamp] = tunique ;
      hash_table->area_stamp[new_stamp] = munique ;
   }
   (*stamp) = new_stamp ;
/*
 if (PARTI_mynode == 0) {
   for (i = 0; i < hash_table->no_buckets; i++)
     printf ("%d\t%d\t%d\n", i, hash_table->cells[i].offset, hash_table->cells[i].stamp);
 }
*/
   
#ifdef  DEBUG
PARTI_gsync();
#endif
}

/*======================================================================*/

int  biggest_stamp_in_mask( mask )
int mask;
{
   int i,j = 1;
   int tmp=0;

   for ( i=FIRST_STAMP; i< MAX_STAMPS && j <= mask; j<<=1,tmp=i, i++ );
   return tmp ;
}
/*======================================================================*/

int   make_mask( num )
int   num ;
{
    int test=1;
    assert( num < 32 && num >= 0);
    test = (test << num);
    return (test );
}
/*======================================================================*/

int make_stamp ( mask )
int  mask ;
{
int i  = 0;
int    test = 1;

 while  ( test !=  mask ) { 
   i++;  test = 1 << i ;
 }

 if ( test != mask ) { 
  return -1 ;
 }

 return i;
}

   
/*======================================================================*/

void read_mask( mask, list)
int    mask, *list;
{
     int    bit = 1;
     int      i = 0;
 
     while( mask ) { 
        list[i++] =  (mask & bit) ;
        mask      =  mask >> 1;
     }
}

/*======================================================================*/

void  print_hash_table( hash_table )
HASHSTR *hash_table;
{

int i,j,n,b;
HASHCELL *buckets,*hashcell;
int *chain, *offset, *stmp;

 PARTI_gsync();
 if ( PARTI_mynode == 0 ) printf("\n ==================================");
 PARTI_gsync();

 for( i=0; i<hash_table->no_buckets;i++ ){ 
    for ( hashcell = &(hash_table->buckets[i]) ; 
        hashcell != NULL;  hashcell = hashcell->next ) {
        chain    = hashcell->key;
        offset   = hashcell->offset;
        stmp    = hashcell->stamp;
        for( j = 0; j< hashcell->tag; j++ ) {
           printf("\n [%d] bucket (%d)-(%d:%d) stmp(%d) tag(%d)", PARTI_mynode ,i, chain[j],offset[j], stmp[j], hashcell->tag);
        }
    }
 }
 PARTI_gsync();

}
/*======================================================================*/

void   clear_mask ( hash_table , mask )
HASHSTR *hash_table ;
int      mask ;
{

 int      no_buckets,  i,j, siz, k, max;
 int      *key, *proc, *local,  *stamp, *offset;
 HASHCELL *bukets, *hashcell, *back, *topcell;

 assert (  CHAINSIZE == 1);
 no_buckets   = hash_table->no_buckets ;
 siz = 0;

 if (hash_table->type == DEFHASH) {
   for( i=0; i < no_buckets; i++ ) {
     topcell =  &(hash_table->buckets[i]); 
     for ( back =  hashcell = topcell;
           hashcell != NULL ; back = hashcell, hashcell = hashcell->next ) {
       stamp  = hashcell->stamp ;
       if (  mask ==  stamp[0] ) {
	 if ( hashcell != topcell ) {
           back->next = hashcell->next ;
           free( hashcell) ;
           hashcell = back; 
	 }
	 else {
	   topcell->key[0]   = -1;
	   topcell->stamp[0] = -1;
           topcell->tag = 0;
         } 
       }
       else if ( mask  &  stamp[0] ) {
            stamp[0] &= ( ~mask) ;
       }
     }
   }
 } 
 else if (hash_table->type == REPHASH) {
   int newmask = ~mask;
   REPHCELL *cells = hash_table->cells;
   for (i = 0; i < no_buckets; i++) {
     cells[i].stamp &= newmask;
   }
 }

 hash_table->no_stamps-- ;
 return; 
}
/*====================================================================*/
